"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Command = void 0;
const node_toolkit_1 = require("@tech_query/node-toolkit");
const parser_1 = require("./parser");
const creator_1 = require("./creator");
const PresetOption = {
    version: { shortcut: 'v', description: 'show Version number' },
    help: { shortcut: 'h', description: 'show Help information' }
};
class Command {
    constructor(meta) {
        this.name = '';
        this.parameters = '';
        this.description = '';
        this.version = '';
        this.options = {};
        this.children = [];
        Object.assign(this, meta);
        for (const command of this.children)
            command.parent = this;
        this.addPreset();
    }
    static nameOf(meta, commandPath) {
        var _a;
        if (typeof meta.bin != 'object')
            return meta.name;
        commandPath = commandPath.replaceAll('\\', '/');
        return (_a = Object.entries(meta.bin).find(([name, path]) => commandPath.endsWith(path.replace(/^\.\//, '')))) === null || _a === void 0 ? void 0 : _a[0];
    }
    static execute(command, args) {
        const { data, options } = (0, parser_1.parseArguments)(args);
        if (!command.parent &&
            (!command.name || !command.version || !command.description)) {
            const [_, commandPath] = process.argv;
            const { meta } = (0, node_toolkit_1.packageOf)(commandPath);
            command.name || (command.name = this.nameOf(meta, commandPath) || '');
            command.description || (command.description = meta.description || '');
            if ((command.version || (command.version = meta.version || '')))
                command.options.version = PresetOption.version;
        }
        command.execute(options, ...data);
    }
    execute(options, ...data) {
        const command = this.children.find(({ name }) => name === data[0]);
        if (command instanceof Command)
            return command.execute(options, ...data.slice(1));
        options = this.checkPattern(this.replaceShortcut(options));
        if ('version' in options) {
            console.log(this.version);
        }
        else if ('help' in options) {
            this.showHelp();
        }
        else if (this.executor instanceof Function) {
            this.executor(options, ...data);
        }
        else {
            throw ReferenceError(`Unknown "${data[0]}" command`);
        }
    }
    replaceShortcut(options) {
        const map = Object.fromEntries(Object.entries(this.options).map(([key, { shortcut }]) => [shortcut, key])), data = {};
        for (const key in options)
            if (key in map)
                data[map[key]] = options[key];
            else
                data[key] = options[key];
        return data;
    }
    checkPattern(options) {
        var _a;
        for (const key in options) {
            const option = this.options[key];
            if (!option)
                throw ReferenceError(`Unknown "${key}" option`);
            if (((_a = option.pattern) === null || _a === void 0 ? void 0 : _a.test(options[key] + '')) === false)
                throw SyntaxError(`"${key}=${options[key]}" doesn't match ${option.pattern}`);
        }
        return options;
    }
    addPreset() {
        const { name, options, children } = this, { version, help } = PresetOption;
        Object.assign(this.options, Object.assign(Object.assign(Object.assign({}, (this.version ? { version } : {})), { help }), options));
        if (name !== 'help' && !children.find(({ name }) => name === 'help'))
            children.push(new Command(Object.assign(Object.assign({ name: 'help' }, help), { parameters: '[command]', executor: (_, command) => this.showHelp(command) })));
    }
    showHelp(command) {
        if (!command)
            return console.log(this + '');
        const that = this.children.find(({ name }) => name === command);
        if (that)
            console.log(that + '');
    }
    *getParentNames() {
        let that = this;
        while ((that = that.parent))
            yield that.name;
    }
    toString() {
        const result = [
            [this.parameters, this.name, ...this.getParentNames()]
                .reverse()
                .join(' ')
                .trim(),
            this.description,
            'Options:\n' + this.toOptionString()
        ];
        if (this.children[0])
            result.push('Commands:\n' + this.toChildrenString());
        return result.join('\n\n');
    }
    toOptionString() {
        return (0, creator_1.createTable)(Object.entries(this.options)
            .sort(([A], [B]) => A.localeCompare(B))
            .map(([name, { shortcut, parameters = '', description = '' }]) => [
            '',
            `${shortcut ? `-${shortcut}, ` : ''}${name[1] ? '-' : ''}-${name}`,
            parameters,
            description
        ]));
    }
    toChildrenString() {
        return (0, creator_1.createTable)(this.children
            .sort(({ name: A }, { name: B }) => A.localeCompare(B))
            .map(({ name, parameters = '', description = '' }) => [
            '',
            name,
            parameters,
            description
        ]));
    }
}
exports.Command = Command;
